(*                       Predecessors
   Lambda-calculus normalizer and other interpreters to
   express and run all the examples in the paper
*)

(* 
#use "pred.ml";;
*)

(* ------------------------------------------------------------------------ 
 *                      Basic Definitions
 *)

(* Untyped pure lambda-calculus, using HOAS 
   All the terms in the paper are closed; correspondingly, we don't
   provide any facilities to express free variables.

   This is not a limitation: an open term can be closed by wrapping it
   into lambda. It still can be normalized since our normalizer
   reduces under lambdas (it follows the normal-order reductions).
*)

type name = string

module type Lam = sig
  type repr
  (* the first argument is the name to give to the bound variable.
     It is useful when printing the term
  *)
  val lam : name -> (repr -> repr) -> repr
  val (/) : repr -> repr -> repr
end

(* Sample terms *)
module Basic(L:Lam) = struct
  open L

  let id = lam "x" (fun x -> x)
  let cmp = lam "f" @@ fun f -> lam "g" @@ fun g -> 
    lam "x" @@ fun x -> f / (g / x)
  let delta = lam "x" (fun x -> x / x)
  let omega = delta / delta
  let fix   = lam "f" @@ fun f -> delta / (lam"u" @@ fun u -> f / (u / u))

  (* Booleans *)
  let btrue  = lam "x" (fun x -> lam "y" (fun y -> x))
  let bfalse = lam "x" (fun x -> lam "y" (fun y -> y))
  let band = lam "p" (fun p -> lam "q" (fun q -> p / q / bfalse))
  let bor  = lam "p" (fun p -> lam "q" (fun q -> p / btrue / q))
  let bnot = lam "p" (fun p -> p / bfalse / btrue)

  (* pairs *)
  let pair = lam "x" @@ fun x -> lam "y" @@ fun y -> lam "p" @@ fun p ->
    p / x / y
  let fst = lam "p" @@ fun p -> p / btrue
  let snd = lam "p" @@ fun p -> p / bfalse

  (* Church numerals *)  
  let c0 = lam "f" @@ fun f -> lam "x" @@ fun x -> x   (* zero, aka bfalse *)
  let c1 = lam "f" @@ fun f -> lam "x" @@ fun x -> f / x
  let succ = lam "n" @@ fun n -> lam "f" @@ fun f -> lam "x" @@ fun x ->
    f / (n / f / x)
  let c2 = succ / c1
  let c3 = succ / c2
  let add = lam "n" @@ fun n -> lam "m" @@ fun m -> 
    lam "f" @@ fun f -> lam "x" @@ fun x -> n / f / (m / f / x)

  let rec ntimes n f x = if n = 0 then x else f / (ntimes (n-1) f x)
  let c99  = lam "f" @@ fun f -> lam "x" @@ fun x -> ntimes 99 f x
  let c100 = lam "f" @@ fun f -> lam "x" @@ fun x -> ntimes 100 f x
end

(* Computing the size of a term *)
module SizeLam = struct
  type repr = int
  let lam: name -> (repr -> repr) -> repr = fun _ f ->
    f 1 + 1
  let (/) : repr -> repr -> repr = fun e1 e2 -> e1 + e2 + 1
end

let (2,3,10,8) = 
  let module M = Basic(SizeLam) in
  (M.id,M.btrue,M.bnot,M.pair)

(* Our second implementation of the Lam signature interprets Lam terms 
   as strings, so we can show them.

   Actually, not quite strings: we need a bit of _context_:
   the precedence and the list of variable names already bound in the context.
   The list lets us detect the name clash and generate a unique name,
   by appending an integer to the variable name. (We try to avoid the renaming
   and keep the user-given names whenever possible.) For that reason, we
   actually maintain a map, which associates to each variable name the counter
   of its distinctly bound occurrences.
*)

module Show = struct
  module VarBoundOccur = 
    Stdlib.Map.Make(struct type t = name let compare = String.compare end)
   type boundvars = int VarBoundOccur.t
   type prec = int
   type repr = boundvars -> prec -> string

   let paren = function 
   | true  -> fun text -> "(" ^ text ^ ")"
   | false -> fun text -> text

   let lam name fbody = 
     fun v p ->
       let updatef = function None -> Some 1 | Some x -> Some (x+1) in
       let v' = VarBoundOccur.update name updatef v in
       let occ = VarBoundOccur.find name v' in
       let name' = if occ = 1 then name else name ^ string_of_int (occ-1) in
       let body = fbody (fun _ _ -> name') v' 0 in
       paren (p > 0) ("L" ^ name' ^ ". " ^ body)

   let (/) e1 e2 = fun v p -> paren (p>10) (e1 v 10 ^ " " ^ e2 v 11)

   let show e = e VarBoundOccur.empty 0
end

let "Lx. x" = 
  let module M = Basic(Show) in
  Show.show M.id
;;

let "Lx. Ly. x" = 
  let module M = Basic(Show) in
  Show.show M.btrue

let "Lp. Lq. p q (Lx. Ly. y)" = 
  let module M = Basic(Show) in
  Show.show M.band

let "Lp. p (Lx. Ly. y) (Lx. Ly. x)" = 
  let module M = Basic(Show) in
  Show.show M.bnot

let "Lx. Ly. Lp. p x y" = 
  let module M = Basic(Show) in
  Show.show M.pair

(* A variant that displays a term as Scheme expression *)
module ShowScheme = struct
  module VarBoundOccur = 
    Stdlib.Map.Make(struct type t = name let compare = String.compare end)
   type boundvars = int VarBoundOccur.t
   type repr = boundvars ->  string

   let paren text = "(" ^ text ^ ")"

   let lam name fbody = 
     fun v ->
       let updatef = function None -> Some 1 | Some x -> Some (x+1) in
       let v' = VarBoundOccur.update name updatef v in
       let occ = VarBoundOccur.find name v' in
       let name' = if occ = 1 then name else name ^ string_of_int (occ-1) in
       let body = fbody (fun _ -> name') v' in
       paren ("lambda" ^ " " ^ paren name' ^ " " ^ body)

   let (/) e1 e2 = fun v -> paren (e1 v ^ " " ^ e2 v)

   let show e = e VarBoundOccur.empty
end

let "(lambda (x) x)" = 
  let module M = Basic(ShowScheme) in
  ShowScheme.show M.id
;;

let "(lambda (x) (lambda (y) x))" = 
  let module M = Basic(ShowScheme) in
  ShowScheme.show M.btrue

let "(lambda (p) (lambda (q) ((p q) (lambda (x) (lambda (y) y)))))" = 
  let module M = Basic(ShowScheme) in
  ShowScheme.show M.band

let "(lambda (p) ((p (lambda (x) (lambda (y) y))) (lambda (x) (lambda (y) x))))"
 = 
  let module M = Basic(ShowScheme) in
  ShowScheme.show M.bnot

let "(lambda (x) (lambda (y) (lambda (p) ((p x) y))))" = 
  let module M = Basic(ShowScheme) in
  ShowScheme.show M.pair

(* Finally, the normalizer, using the normal-order reductions
 
   It is the term transformer, to another Lam implementation (which is
   often Show). Some normalization may occur only upon `readout'
   (printing the term in Show), because of dyn that occurs within
   HOAS lambdas.

  Thanks to HOAS we do not have to worry about variable capture here.
  The name conflicts could occur when we print out a term; but the
  Show module can deal with these conflicts.
*)

module Norm(L:Lam) = struct
  type repr = 
    | Lam of name * (repr -> repr)      (* lambda-abstraction *)
    | Neu of L.repr                     (* Neutral term *)
    | Thu of (unit -> repr)             (* Delayed substitution *)

  (* This function performs all delayed substitutions *)
  let rec dyn : repr -> L.repr = function
    | Neu x     -> x
    | Lam (n,f) -> L.lam n (fun x -> dyn @@ f (Neu x))
    | Thu th    -> dyn @@ th ()

  let rec (/) e1 e2 = match e1 with
    (* If we found a redex, we delay it for now since we don't
       yet know if this the leftmost outermost one.
       The function dyn will know, and force the thunk 
     *)
  | Lam (n,f) -> Thu (fun () -> f e2)   (* beta-redex *)
  | Thu th    -> Thu (fun () -> th () / e2)
  | Neu e     -> Neu L.(e / dyn e2)

  let lam name fbody = Lam (name,fbody)
end

module NormShow = Norm(Show)
;;

let nshow e = Show.show (NormShow.dyn e)

let _ = 
  let module M = Basic(NormShow) in
  nshow M.bnot

let true = 
  let module M = Basic(NormShow) in
  let open NormShow in
  nshow M.(bnot / btrue) = nshow M.bfalse

let true = 
  let module M = Basic(NormShow) in
  let open NormShow in
  nshow M.(bnot / bfalse) = nshow M.btrue

let true = 
  let module M = Basic(NormShow) in
  let open NormShow in
  nshow M.(lam "z" @@ fun z -> band / btrue / z) = nshow (lam "z" (fun z -> z)) 

(* checking that this is really a normal-order evaluation *)
let true = 
  let module M = Basic(NormShow) in
  let open NormShow in
  nshow M.(band / bfalse / omega) = nshow M.bfalse

let true = 
  let module M = Basic(NormShow) in
  let open NormShow in
  nshow M.(lam "x" @@ fun x -> lam "y" @@ fun y -> 
    fst / (pair / x / y)) = 
  nshow (lam "x" @@ fun x -> lam "y" @@ fun y -> x)

let true = 
  let module M = Basic(NormShow) in
  let open NormShow in
  nshow M.(lam "x" @@ fun x -> lam "y" @@ fun y -> 
    snd / (pair / x / y)) = 
  nshow (lam "x" @@ fun x -> lam "y" @@ fun y -> y)

(* Norm with counting beta-reductions *)
module NormCount(L:Lam) = struct
  include Norm(L)
  let count = ref 0

  let rec (/) e1 e2 = match e1 with
    (* If we found a redex, we delay it for now since we don't
       yet know if this the leftmost outermost one.
       The function dyn will know, and force the thunk 
     *)
  | Lam (n,f) -> Thu (fun () -> incr count; f e2)   (* beta-redex *)
  | Thu th    -> Thu (fun () -> th () / e2)
  | Neu e     -> Neu L.(e / dyn e2)
end

module NormCShow = NormCount(Show)
;;

let nshow e = Show.show (NormCShow.dyn e)

let withcount th = 
  let () = NormCShow.count := 0 in
  let r = th () in
  (r,!NormCShow.count)

let ("Lp. p (Lx. Ly. y) (Lx. Ly. x)", 0) = 
  let module M = Basic(NormCShow) in
  withcount @@ fun () -> nshow M.bnot

let (true,3) = 
  let module M = Basic(NormCShow) in
  let open NormCShow in
  withcount @@ fun () -> 
    nshow M.(bnot / btrue) = nshow M.bfalse

let (true,3) = 
  let module M = Basic(NormCShow) in
  let open NormCShow in
  withcount @@ fun () -> 
    nshow M.(bnot / bfalse) = nshow M.btrue

let (true,4) = 
  let module M = Basic(NormCShow) in
  let open NormCShow in
  withcount @@ fun () -> 
    nshow M.(lam "z" @@ fun z -> band / btrue / z) = 
    nshow (lam "z" (fun z -> z)) 

(* checking that this is really a normal-order evaluation *)
let (true,4) = 
  let module M = Basic(NormCShow) in
  let open NormCShow in
  withcount @@ fun () -> 
    nshow M.(band / bfalse / omega) = nshow M.bfalse

let (true,6) = 
  let module M = Basic(NormCShow) in
  let open NormCShow in
  withcount @@ fun () -> 
    nshow M.(lam "x" @@ fun x -> lam "y" @@ fun y -> 
      fst / (pair / x / y)) = 
    nshow (lam "x" @@ fun x -> lam "y" @@ fun y -> x)

let (true,6) = 
  let module M = Basic(NormCShow) in
  let open NormCShow in
  withcount @@ fun () -> 
    nshow M.(lam "x" @@ fun x -> lam "y" @@ fun y -> 
      snd / (pair / x / y)) = 
    nshow (lam "x" @@ fun x -> lam "y" @@ fun y -> y)

(* Successor is the constant-time operation *)
let (true,3) = 
  let module M = Basic(NormCShow) in
  let open NormCShow in
  withcount @@ fun () -> 
    nshow M.(succ / c0) =
    nshow M.c1

let (true,3) = 
  let module M = Basic(NormCShow) in
  let open NormCShow in
  let s2 = nshow M.c2 in
  withcount @@ fun () -> 
    nshow M.(succ / c1) = s2

let (true,3) = 
  let module M = Basic(NormCShow) in
  let open NormCShow in
  withcount @@ fun () -> 
    nshow M.(succ / c99) =
    nshow M.c100

(*------------------------------------------------------------------------ 
 * Predecesors 
 *)

(* The Kleene predecessor *)

let _ =
  let module P(L:Lam) = struct 
    include Basic(L)
    open L
    let p0 = pair / c0 / c0 
    let supp = lam "p" @@ fun p -> pair / (snd / p) / (succ / (snd / p))
    let pred = lam "n" @@ fun n -> fst / (n / supp / p0)
    (* for checking correctness *)    
    let rfl = fst
    let h   = cmp / rfl / supp
    end 
  in
  let 41 =
    let module NSize = Norm(SizeLam) in let module M = P(NSize) in 
    NSize.dyn @@ M.pred in
  let s =
    let module NScheme = Norm(ShowScheme) in let module M = P(NScheme) in 
    ShowScheme.show @@ NScheme.dyn @@ M.pred in
  let pred = 
    let module NNormCShow = Norm(NormCShow) in
    let module M = P(NNormCShow) in NNormCShow.dyn @@ M.pred in
  let  "Ln. n (Lp. Lp1. p1 (p (Lx. Ly. y)) (Lf. Lx. f (p (Lx1. Ly. y) f x))) (Lp. p (Lf. Lx. x) (Lf. Lx. x)) (Lx. Ly. x)" =
    nshow pred in
  let module M = Basic(NormCShow) in 
  let open NormCShow in let open M in
  let (true,10) =
    withcount @@ fun () -> 
      nshow @@ pred / c1 = nshow c0 in
  let (true, 604) =
    withcount @@ fun () -> 
      nshow @@ pred / c100 = nshow c99 in
   (* correctness *)
   let module MP = P(NormCShow) in
   let () = assert (nshow MP.(h / p0) = nshow c0) in
   let () = assert MP.(nshow (cmp / h / supp) = nshow (cmp / succ / h)) in
  (s,MP.(nshow (cmp / h / supp)))

(* More predecessors, generally *)

(* int option predecessor *)

let _ =
  let module P(L:Lam) = struct 
    include Basic(L)
    open L
    let p0 = lam "k" @@ fun k -> lam "y" @@ fun y -> y (* None *)
    let supp = lam "p" @@ fun p -> 
      lam "k" @@ fun k -> lam "y" @@ fun y ->
        k / (p / succ / c0)
    let pred = lam "n" @@ fun n -> (n / supp / p0) / id / c0
    let rfl = lam "p" @@ fun p -> p / id / c0
    let h   = cmp / rfl / supp
    end 
  in
  let 35 =
    let module NSize = Norm(SizeLam) in let module M = P(NSize) in 
    NSize.dyn @@ M.pred in
  let s =
    let module NScheme = Norm(ShowScheme) in let module M = P(NScheme) in 
    ShowScheme.show @@ NScheme.dyn @@ M.pred in
  let pred = 
    let module NNormCShow = Norm(NormCShow) in
    let module M = P(NNormCShow) in NNormCShow.dyn @@ M.pred in
  let  "Ln. n (Lp. Lk. Ly. k (p (Ln1. Lf. Lx. f (n1 f x)) (Lf. Lx. x))) (Lk. Ly. y) (Lx. x) (Lf. Lx. x)" =
    nshow pred in
  let module M = Basic(NormCShow) in 
  let open NormCShow in let open M in
  let (true, 9) =
    withcount @@ fun () -> 
      nshow @@ pred / c1 = nshow c0 in
  let (true, 603) =
    withcount @@ fun () -> 
      nshow @@ pred / c100 = nshow c99 in
   let module MP = P(NormCShow) in
   let () = assert (nshow MP.(h / p0) = nshow c0) in
   let () = assert MP.(nshow (cmp / h / supp) = nshow (cmp / succ / h)) in
  (s)

(* optimized one *)
let _ =
  let module P(L:Lam) = struct 
    include Basic(L)
    open L
    let p0 f x = lam "k" @@ fun k -> x
    let supp f x = lam "p" @@ fun p -> lam "k" @@ fun k ->
      k / (p / f)
    let pred = lam "n" @@ fun n -> 
      lam "f" @@ fun f -> lam "x" @@ fun x -> 
        (n / supp f x / p0 f x) / id
    (* for checking correctness *)    
    let rfl    = lam "p" @@ fun p -> p / id
    let h f x  = cmp / rfl / supp f x
    let bindfx th = lam "f" @@ fun f -> lam "x" @@ fun x -> th f x
  end 
  in
  let 18 =
    let module NSize = Norm(SizeLam) in let module M = P(NSize) in 
    NSize.dyn @@ M.pred in
  let s =
    let module NScheme = Norm(ShowScheme) in let module M = P(NScheme) in 
    ShowScheme.show @@ NScheme.dyn @@ M.pred in
  let pred = 
    let module NNormCShow = Norm(NormCShow) in
    let module M = P(NNormCShow) in NNormCShow.dyn @@ M.pred in
  let  "Ln. Lf. Lx. n (Lp. Lk. k (p f)) (Lk. x) (Lx1. x1)" =
    nshow pred in
  let module M = Basic(NormCShow) in 
  let open NormCShow in let open M in
  let (true, 7) =
    withcount @@ fun () -> 
      nshow @@ pred / c1 = nshow c0 in
  let (true, 205) =
    withcount @@ fun () -> 
      nshow @@ pred / c100 = nshow c99 in
   let module MP = P(NormCShow) in
   let () = assert (nshow MP.(bindfx (fun f x -> h f x / p0 f x)) = 
                    nshow MP.(bindfx (fun f x -> c0 / f / x))) in
   let () = assert MP.(nshow (bindfx (fun f x -> cmp / h f x / supp f x)) = 
                       nshow (bindfx (fun f x -> cmp / f / h f x))) in
  (s)



(* More predecessors, specifically *)

let _ =
  let module P(L:Lam) = struct 
    include Basic(L)
    open L
    let p0 = lam "f" @@ fun f -> lam "x" @@ fun x -> btrue
    let supp = lam "p" @@ fun p -> 
      p / (lam "z" @@ fun z -> bfalse) / bfalse / c0 / (succ / p)
    let pred = lam "n" @@ fun n -> n / supp / p0
    end 
  in
  let 34 =
    let module NSize = Norm(SizeLam) in let module M = P(NSize) in 
    NSize.dyn @@ M.pred in
  let s =
    let module NScheme = Norm(ShowScheme) in let module M = P(NScheme) in 
    ShowScheme.show @@ NScheme.dyn @@ M.pred in
  let pred = 
    let module NNormCShow = Norm(NormCShow) in
    let module M = P(NNormCShow) in NNormCShow.dyn @@ M.pred in
  let  "Ln. n (Lp. p (Lz. Lx. Ly. y) (Lx. Ly. y) (Lf. Lx. x) (Lf. Lx. f (p f x))) (Lf. Lx. Lx1. Ly. x1)" =
    nshow pred in
  let module M = Basic(NormCShow) in 
  let open NormCShow in let open M in
  let (true, 8) =
    withcount @@ fun () -> 
      nshow @@ pred / c1 = nshow c0 in
  let (true, 30302) =
    withcount @@ fun () -> 
      nshow @@ pred / c100 = nshow c99 in
  (s)

(* The one from 1992 *)
let _ =
  let module P(L:Lam) = struct 
    include Basic(L)
    open L
    let p0 = lam "f" @@ fun f -> lam "x" @@ fun x -> c0
    let supp = lam "p" @@ fun p -> 
      p / id / (succ / p)
    let pred = lam "n" @@ fun n -> n / supp / p0
    end
  in
  let 24 =
    let module NSize = Norm(SizeLam) in let module M = P(NSize) in 
    NSize.dyn @@ M.pred in
  let s =
    let module NScheme = Norm(ShowScheme) in let module M = P(NScheme) in 
    ShowScheme.show @@ NScheme.dyn @@ M.pred in
  let pred = 
    let module NNormCShow = Norm(NormCShow) in
    let module M = P(NNormCShow) in NNormCShow.dyn @@ M.pred in
  let  "Ln. n (Lp. p (Lx. x) (Lf. Lx. f (p f x))) (Lf. Lx. Lf1. Lx1. x1)" =
    nshow pred in
  let module M = Basic(NormCShow) in 
  let open NormCShow in let open M in
  let (true, 6) =
    withcount @@ fun () -> 
      nshow @@ pred / c1 = nshow c0 in
  let (true, 14) =
    let c2  = lam "f" @@ fun f -> lam "x" @@ fun x -> ntimes 2 f x in
    withcount @@ fun () -> 
      nshow @@ pred / c2 = nshow c1 in
  let (true, 31) =
    let c3  = lam "f" @@ fun f -> lam "x" @@ fun x -> ntimes 3 f x in
    let c2  = lam "f" @@ fun f -> lam "x" @@ fun x -> ntimes 2 f x in
    withcount @@ fun () -> 
      nshow @@ pred / c3 = nshow c2 in
  let (true, 66) =
    let c4  = lam "f" @@ fun f -> lam "x" @@ fun x -> ntimes 4 f x in
    let c3  = lam "f" @@ fun f -> lam "x" @@ fun x -> ntimes 3 f x in
    withcount @@ fun () -> 
      nshow @@ pred / c4 = nshow c3 in
  let (true, 137) =
    let c5  = lam "f" @@ fun f -> lam "x" @@ fun x -> ntimes 5 f x in
    let c4  = lam "f" @@ fun f -> lam "x" @@ fun x -> ntimes 4 f x in
    withcount @@ fun () -> 
      nshow @@ pred / c5 = nshow c4 in
  (* could not finish in 5 minutes
  let r =
    withcount @@ fun () -> 
      nshow @@ pred / c100 = nshow c99 in
  *)
  (s)

(* a bit more optimized *)
let _ =
  let module P(L:Lam) = struct 
    include Basic(L)
    open L
    let p0 = lam "f" @@ fun f -> lam "x" @@ fun x -> c0
    let c1 = id
    let supp = lam "p" @@ fun p -> 
      p / (lam "z" @@ fun z -> succ / p) / c1
    let pred = lam "n" @@ fun n -> n / supp / p0
    end 
  in
  let 25 =
    let module NSize = Norm(SizeLam) in let module M = P(NSize) in 
    NSize.dyn @@ M.pred in
  let s =
    let module NScheme = Norm(ShowScheme) in let module M = P(NScheme) in 
    ShowScheme.show @@ NScheme.dyn @@ M.pred in
  let pred = 
    let module NNormCShow = Norm(NormCShow) in
    let module M = P(NNormCShow) in NNormCShow.dyn @@ M.pred in
  let "Ln. n (Lp. p (Lz. Lf. Lx. f (p f x)) (Lx. x)) (Lf. Lx. Lf1. Lx1. x1)" =
    nshow pred in
  let module M = Basic(NormCShow) in 
  let open NormCShow in let open M in
  let (true, 6) =
    withcount @@ fun () -> 
      nshow @@ pred / c1 = nshow c0 in
  let (true, 20098) =
    withcount @@ fun () -> 
      nshow @@ pred / c100 = nshow c99 in
  (s)

(*
let _ =
  let module P(L:Lam) = struct 
    include Basic(L)
    open L
    let p0 = lam "f" @@ fun f -> c0
    let supp = lam "c" @@ fun c -> c / succ
    let pred = lam "n" @@ fun n -> n / supp / p0
    end 
  in
  let l =
    let module NSize = Norm(SizeLam) in let module M = P(NSize) in 
    NSize.dyn @@ M.pred in
  let s =
    let module NScheme = Norm(ShowScheme) in let module M = P(NScheme) in 
    ShowScheme.show @@ NScheme.dyn @@ M.pred in
  let pred = 
    let module NNormCShow = Norm(NormCShow) in
    let module M = P(NNormCShow) in NNormCShow.dyn @@ M.pred in
  let  s1 =
    nshow pred in
  let module M = Basic(NormCShow) in 
  let open NormCShow in let open M in
  let (true, r1) =
    withcount @@ fun () -> 
      nshow @@ pred / c1 = nshow c0 in
  let (true, r2) =
    withcount @@ fun () -> 
      nshow @@ pred / c100 = nshow c99 in
  (s,l,s1,r1,r2)
*)

(* Even more optimized and inspirational, using ( * ) *)
let _ =
  let module P(L:Lam) = struct 
    include Basic(L)
    open L
    let p0 = lam "f" @@ fun f -> lam "x" @@ fun x -> c0
    let c1 = id
    let supp = lam "c" @@ fun c -> c / succ / c1
    let pred = lam "n" @@ fun n -> n / supp / p0
    end 
  in
  let 25 =
    let module NSize = Norm(SizeLam) in let module M = P(NSize) in 
    NSize.dyn @@ M.pred in
  let s =
    let module NScheme = Norm(ShowScheme) in let module M = P(NScheme) in 
    ShowScheme.show @@ NScheme.dyn @@ M.pred in
  let pred = 
    let module NNormCShow = Norm(NormCShow) in
    let module M = P(NNormCShow) in NNormCShow.dyn @@ M.pred in
  let  "Ln. n (Lc. c (Ln1. Lf. Lx. f (n1 f x)) (Lx. x)) (Lf. Lx. Lf1. Lx1. x1)" =
    nshow pred in
  let module M = Basic(NormCShow) in 
  let open NormCShow in let open M in
  let (true, 6) =
    withcount @@ fun () -> 
      nshow @@ pred / c1 = nshow c0 in
  let (true, 14757) =
    withcount @@ fun () -> 
      nshow @@ pred / c100 = nshow c99 in
  (s)

(* a bit yet more optimized *)
let _ =
  let module P(L:Lam) = struct 
    include Basic(L)
    open L
    let p0 = lam "f" @@ fun f -> lam "x" @@ fun x -> id
    let supp = lam "p" @@ fun p -> 
      lam "f" @@ fun f ->
        p / (lam "g" @@ fun g -> lam "x" @@ fun x -> f / (g / x)) / f 
    let pred = lam "n" @@ fun n -> n / supp / p0
    end 
  in
  let 21 =
    let module NSize = Norm(SizeLam) in let module M = P(NSize) in 
    NSize.dyn @@ M.pred in
  let s =
    let module NScheme = Norm(ShowScheme) in let module M = P(NScheme) in 
    ShowScheme.show @@ NScheme.dyn @@ M.pred in
  let pred = 
    let module NNormCShow = Norm(NormCShow) in
    let module M = P(NNormCShow) in NNormCShow.dyn @@ M.pred in
  let "Ln. n (Lp. Lf. p (Lg. Lx. f (g x)) f) (Lf. Lx. Lx1. x1)" =
    nshow pred in
  let module M = Basic(NormCShow) in 
  let open NormCShow in let open M in
  let (true, r1) =
    withcount @@ fun () -> 
      nshow @@ pred / c1 = nshow c0 in
  let (true, 9906) =
    withcount @@ fun () -> 
      nshow @@ pred / c100 = nshow c99 in
  (s)



(* Un-application *)
(* Because of the fix, we can't normalize *)
let _ =
  let module P(L:Lam) = struct 
    include Basic(L)
    open L
    let p0 = lam "k" @@ fun k -> lam "y" @@ fun y -> y (* None *)
    let reif = lam "n" @@ fun n ->
      n / (lam "a" @@ fun a -> lam "k" @@ fun k -> lam "y" @@ fun y -> k / a) /
      p0
    let refl f x = 
      fix / (lam "s" @@ fun s -> lam "p" @@ fun p -> 
        p / (lam "q" @@ fun q -> f / (s / q)) / x)
    let predp = lam "p" @@ fun p -> p / id / p0
    let pred = lam "n" @@ fun n -> lam "f" @@ fun f -> lam "x" @@ fun x ->
      refl f x / (predp / (reif / n))
  end
  in
  let 54 =
    let module M = P(SizeLam) in M.pred in
  let "Ln. Lf. Lx. (Lf1. (Lx1. x1 x1) (Lu. f1 (u u))) (Ls. Lp. p (Lq. f (s q)) x) ((Lp. p (Lx1. x1) (Lk. Ly. y)) ((Ln1. n1 (La. Lk. Ly. k a) (Lk. Ly. y)) n))"
      = 
    let module M = P(Show) in 
    Show.show M.pred in
  let module M = P(NormCShow) in 
  let open NormCShow in let open M in
  let (true, 16) =
    withcount @@ fun () -> 
      nshow @@ pred / c1 = nshow c0 in 
  let (true, 709) =
    withcount @@ fun () -> 
      nshow @@ pred / c100 = nshow c99 in
  ()

let _ =
  let module P(L:Lam) = struct 
    include Basic(L)
    open L
    let p0 x = lam "k" @@ fun k -> x (* None *)
    let reif x = lam "n" @@ fun n ->
      n / (lam "a" @@ fun a -> lam "k" @@ fun k -> k / a) /
      p0 x
    let refl f = 
      fix / (lam "s" @@ fun s -> lam "p" @@ fun p -> 
        p / (lam "q" @@ fun q -> f / (s / q)))
    let predp = lam "p" @@ fun p -> p / id
    let pred = lam "n" @@ fun n -> lam "f" @@ fun f -> lam "x" @@ fun x ->
      refl f / (predp / (reif x / n))
     (* Manually desugared: because of fix we can't desugar automatically *)
    let pred' = lam "n" @@ fun n -> lam "f" @@ fun f -> lam "x" @@ fun x ->
      fix / 
       (lam "s" @@ fun s -> lam "p" @@ fun p -> 
         p / (lam "q" @@ fun q -> f / (s / q))) /
      (n / (lam "a" @@ fun a -> lam "k" @@ fun k -> k / a) / 
       (lam "k" @@ fun k -> x) / (lam "y" @@ fun y -> y))
  end
  in
  let 40 =
    let module M = P(SizeLam) in M.pred' in
  let "Ln. Lf. Lx. (Lf1. (Lx1. x1 x1) (Lu. f1 (u u))) (Ls. Lp. p (Lq. f (s q))) (n (La. Lk. k a) (Lk. x) (Ly. y))" = 
    let module M = P(Show) in 
    Show.show M.pred' in
  let module M = P(NormCShow) in 
  let open NormCShow in let open M in
  let (true, 12) =
    withcount @@ fun () -> 
      nshow @@ pred' / c1 = nshow c0 in 
  let (true, 606) =
    withcount @@ fun () -> 
      nshow @@ pred' / c100 = nshow c99 in
  ()

(* Normalizable, because of using cn *)
let _ =
  let module P(L:Lam) = struct 
    include Basic(L)
    open L
    let p0 x = lam "k" @@ fun k -> x (* None *)
    let reif x = lam "n" @@ fun n ->
      n / (lam "a" @@ fun a -> lam "k" @@ fun k -> k / a) /
      p0 x
    let refl n f = 
      n / (lam "s" @@ fun s -> lam "p" @@ fun p -> 
        p / (lam "q" @@ fun q -> f / (s / q))) / n
    let predp = lam "p" @@ fun p -> p / id
    let pred = lam "n" @@ fun n -> lam "f" @@ fun f -> lam "x" @@ fun x ->
      refl n f / (predp / (reif x / n))
    end 
  in
  let 31 =
    let module NSize = Norm(SizeLam) in let module M = P(NSize) in 
    NSize.dyn @@ M.pred in
  let s =
    let module NScheme = Norm(ShowScheme) in let module M = P(NScheme) in 
    ShowScheme.show @@ NScheme.dyn @@ M.pred in
  let pred = 
    let module NNormCShow = Norm(NormCShow) in
    let module M = P(NNormCShow) in NNormCShow.dyn @@ M.pred in
  let "Ln. Lf. Lx. n (Ls. Lp. p (Lq. f (s q))) n (n (La. Lk. k a) (Lk. x) (Lx1. x1))" =
    nshow pred in
  let module M = Basic(NormCShow) in 
  let open NormCShow in let open M in
  let ("Lf. Lx. Lx1. x1", 4) =
    withcount @@ fun () -> 
      nshow @@ pred / c0 in
  let (true, 11) =
    withcount @@ fun () -> 
      nshow @@ pred / c1 = nshow c0 in
  let (true, 506) =
    withcount @@ fun () -> 
      nshow @@ pred / c100 = nshow c99 in
  (s)

;;

(*------------------------------------------------------------------------
 * Side remark on typing 
 *)

type nat = {nat: 'a. ('a->'a) -> 'a -> 'a}

let c0 : nat = {nat = fun f -> fun x -> x}
let c1 : nat = {nat = fun f -> fun x -> f x}
let c2 : nat = {nat = fun f -> fun x -> f (f x)}
let succ : nat -> nat = fun {nat} -> {nat = fun f -> fun x -> f (nat f x)}
let show_nat : nat -> int = fun {nat} -> nat Stdlib.succ 0

let 3 = show_nat @@ succ c2 

let pred : nat -> nat = fun {nat} ->
  {nat = fun (type a) -> fun f -> fun x ->
    nat (fun (p : (a -> a) -> a) -> fun k -> k (p f)) 
      (fun k -> x) (fun y -> y)}

let 2 = show_nat @@ (pred @@ succ c2)
let 1 = show_nat @@ (pred @@ c2)
let 0 = show_nat @@ (pred @@ c1)
let 0 = show_nat @@ (pred @@ c0)

(*------------------------------------------------------------------------
 * Predecessors on trees
 *)

module Tree(L:Lam) = struct
  include Basic(L)
  open L

  let leaf = lam "a" @@ fun a -> 
    lam "cl" @@ fun cl -> lam "cn" @@ fun cn -> cl / a
  let node = lam "b1" @@ fun b1 -> lam "b2" @@ fun b2 -> 
    lam "cl" @@ fun cl -> lam "cn" @@ fun cn -> 
      cn / (b1 / cl / cn) / (b2 / cl / cn)

  (* Sample tree *)
  let t3 = node / (node / (leaf / c1) / (leaf / c2)) / (leaf / c3)
  let t4 = node / t3 / (node / (leaf / c1) / (leaf / c2))
end

(* summation of all leaf values *)
let "Lf. Lx. f (f (f (f (f (f x)))))" = 
  let module M = Tree(NormCShow) in
  let open NormCShow in let open M in
  nshow @@ t3 / id / add

(* checking ( * ) for trees *)
let true = 
  let module M = Tree(NormCShow) in
  let open NormCShow in let open M in
  nshow (t3 / leaf / node) = nshow t3

(* left and right, using reflection-reification, with fixpoint *)
let _ =
  let module P(L:Lam) = struct 
    include Tree(L)
    open L
    let pleaf = leaf
    let pnode = lam "b1" @@ fun b1 -> lam "b2" @@ fun b2 -> 
      lam "cl" @@ fun cl -> lam "cn" @@ fun cn -> cn / b1 / b2
    let reif = lam "t" @@ fun t -> t / pleaf / pnode
    let refl = 
      fix / (lam "s" @@ fun s -> lam "t" @@ fun t ->
        t / leaf / (lam "b1" @@ fun b1 -> lam "b2" @@ fun b2 ->
          node / (s / b1) / (s / b2)))
    let pleft  = lam "t" @@ fun t -> t / pleaf / btrue
    let pright = lam "t" @@ fun t -> t / pleaf / bfalse
    let left  = cmp / refl / (cmp / pleft  / reif)
    let right = cmp / refl / (cmp / pright / reif)
    end 
  in
  let module M = P(NormCShow) in 
  let open NormCShow in let open M in
  let true =
    nshow @@ refl / (reif / t3) = nshow t3 in
  let true =
    nshow @@ left / t3  = nshow (node / (leaf / c1) / (leaf / c2)) in
  let true =
    nshow @@ right / t3  = nshow (leaf / c3) in
  let true =
    nshow @@ left / t4  = nshow t3 in
  let true =
    nshow @@ right / t4  = nshow @@ left / t3 in
  ()

(* left and right, using reflection-reification, with fixpoint 
   More optimal version, as in the Un-application section
*)
let _ =
  let module P(L:Lam) = struct 
    include Tree(L)
    open L
    let pleaf cl = lam "a" @@ fun a -> lam "cn" @@ fun cn -> cl / a
    let pnode = lam "b1" @@ fun b1 -> lam "b2" @@ fun b2 -> 
      lam "cn" @@ fun cn -> cn / b1 / b2
    let reif cl = lam "t" @@ fun t -> t / pleaf cl / pnode
    let refl cn = 
      fix / (lam "s" @@ fun s -> lam "t" @@ fun t ->
        t / (lam "b1" @@ fun b1 -> lam "b2" @@ fun b2 ->
                cn / (s / b1) / (s / b2)))
    let pleft  = lam "t" @@ fun t -> t / btrue
    let pright = lam "t" @@ fun t -> t / bfalse
    let left  = lam "t" @@ fun t -> 
      lam "cl" @@ fun cl -> lam "cn" @@ fun cn -> 
        refl cn / (pleft / (reif cl / t))
    let right  = lam "t" @@ fun t -> 
      lam "cl" @@ fun cl -> lam "cn" @@ fun cn -> 
        refl cn / (pright / (reif cl / t))
    end 
  in
  let module M = P(NormCShow) in 
  let open NormCShow in let open M in
  let true =
    nshow @@ left / t3  = nshow (node / (leaf / c1) / (leaf / c2)) in
  let true =
    nshow @@ right / t3  = nshow (leaf / c3) in
  let true =
    nshow @@ left / t4  = nshow t3 in
  let true =
    nshow @@ right / t4  = nshow @@ left / t3 in
  ()

(* left and right, using reflection-reification, without fixpoint *)
let _ =
  let module P(L:Lam) = struct 
    include Tree(L)
    open L
    let pleaf = leaf
    let pnode = lam "b1" @@ fun b1 -> lam "b2" @@ fun b2 -> 
      lam "cl" @@ fun cl -> lam "cn" @@ fun cn -> cn / b1 / b2
    let reif = lam "t" @@ fun t -> t / pleaf / pnode
    let refl = lam "t'" @@ fun t' -> 
      t' / id / (lam "s1" @@ fun s1 -> lam "s2" @@ fun s2 ->
        lam "t" @@ fun t ->
        t / leaf / (lam "b1" @@ fun b1 -> lam "b2" @@ fun b2 ->
          node / (s1 / b1) / (s2 / b2)))
    let pleft  = lam "t" @@ fun t -> t / pleaf / btrue
    let pright = lam "t" @@ fun t -> t / pleaf / bfalse
    let left  = lam "t" @@ fun t -> refl / t / (pleft  / (reif / t))
    let right = lam "t" @@ fun t -> refl / t / (pright / (reif / t))
    end 
  in
  let module M = P(NormCShow) in 
  let open NormCShow in let open M in
  let true =
    nshow @@ left / t3  = nshow (node / (leaf / c1) / (leaf / c2)) in
  let true =
    nshow @@ right / t3  = nshow (leaf / c3) in
  let true =
    nshow @@ left / t4  = nshow t3 in
  let true =
    nshow @@ right / t4  = nshow @@ left / t3 in
  ()

(* left and right, optimal Boehm-Berarducci-like (the analogue of the
   smalles/fastest predecessor)
*)
let _ =
  let module P(L:Lam) = struct 
    include Tree(L)
    open L
    let pleaf cl = lam "a" @@ fun a -> lam "cn'" @@ fun cn' -> cl / a
    let pnode cn = lam "b1" @@ fun b1 -> lam "b2" @@ fun b2 -> 
      lam "cn'" @@ fun cn' -> cn' / (b1 / cn) / (b2 / cn)
    let left = lam "t" @@ fun t ->
      lam "cl" @@ fun cl -> lam "cn" @@ fun cn -> 
        t / (pleaf cl) / (pnode cn) / btrue
    let right = lam "t" @@ fun t ->
      lam "cl" @@ fun cl -> lam "cn" @@ fun cn -> 
        t / (pleaf cl) / (pnode cn) / bfalse
    end 
  in
  let module M = P(NormCShow) in 
  let open NormCShow in let open M in
  let "Lt. Lcl. Lcn. t (La. Lcn'. cl a) (Lb1. Lb2. Lcn'. cn' (b1 cn) (b2 cn)) (Lx. Ly. x)"
     = nshow left in
  let true =
    nshow @@ left / t3  = nshow (node / (leaf / c1) / (leaf / c2)) in
  let true =
    nshow @@ right / t3  = nshow (leaf / c3) in
  let true =
    nshow @@ left / t4  = nshow t3 in
  let true =
    nshow @@ right / t4  = nshow @@ left / t3 in
  ()

;;
